/*!
 * @file  RTC_OmniWheel.cpp
 * @brief 全方位移動搭乗型ロボットプラットフォームOmni-Wheelを制御するためのRTC
 *
 * @author Masaharu Shimizu@Future Robotics Technology Center, Chiba Institute of Technology
 *
 * @date 2009/06/22 IIS.idlフォーマットのTimedVelocityで速度指令、TimedPositionで自己位置出力を追加
 *                  タイムスタンプも重要情報となったので出力するように追加
 *                  共通デモまでのSGIフォーマットとは、x,y軸の扱いが違うので注意が必要
 *                  台車操作クラスCMoveWheelChairで計算される角度オドメトリの正方向が左回りに変更されたためにこちらも変更
 * @date 2009/08/20 旧式のポートはコメントアウト
 * @date 2009/08/27 産総研追従プログラムにてTimedPositionのIDを利用しているとのことなので100を設定
 * @date 2009/09/07 コンフィギュレーションで、ログ出力をON、OFFできるようにする
 * @date 2009/09/07 コンフィギュレーションで、オドメトリをアクティベイトの度にリセットするか選択できるようにする
 * @date 2009/09/07 コンフィギュレーションで、オドメトリのゲイン調整を可能なパラメータを設定する
 * @date 2009/09/07 コンフィギュレーションで、オドメトリ.idを設定できるようにする
 * @date 2009/10/28 OpenRTM1.0RCに対応する
 * @date 2010/06/30 OpenRTM1.0で動作確認済み
 * $Id$
 */

#include "RTC_OmniWheel.h"
#include <unistd.h>
#include <math.h>
#include <iostream>
#include <iomanip>

// Module specification
// <rtc-template block="module_spec">
static const char* rtc_omniwheel_spec[] = {
		"implementation_id",     "RTC_OmniWheel",
		"type_name",             "RTC_OmniWheel",
		"description",           "Control OmniWheel Chair and obtain odometry",
		"version",               "1.0.0",
		"vendor",                "Chiba Institute of Technology",
		"category",              "Control",
		"activity_type",         "SPORADIC",
		"kind",                  "DataFlowComponent",
		"max_instance",          "1",
		"language",              "C++",
		"lang_type",             "compile",
		// Configuration variables
		"conf.default.odo_id", "100",
		"conf.default.odo_pos_x", "0.0",
		"conf.default.odo_pos_y", "0.0",
		"conf.default.odo_pos_theta", "0.0",
		"conf.default.check_ResetPos", "1",
		"conf.default.check_Log", "1",
		"conf.default.gain_odo_pos_x", "1.0",
		"conf.default.gain_odo_pos_y", "1.0",
		"conf.default.gain_odo_pos_theta", "1.0",
		""
};
// </rtc-template>

RTC_OmniWheel::RTC_OmniWheel(RTC::Manager* manager)
 : RTC::DataFlowComponentBase(manager),
 // <rtc-template block="initializer">
 m_inVelIISIn("inVelIIS",m_inVelIIS),
 m_outOdometryIISOut("outOdometryIIS",m_outOdometryIIS)

// </rtc-template>
{
	// Registration: InPort/OutPort/Service
	// <rtc-template block="registration">
	// Set InPort buffers
	registerInPort("inVelIIS", m_inVelIISIn);

	// Set OutPort buffer
	registerOutPort("outOdometryIIS", m_outOdometryIISOut);

	// Set service provider to Ports

	// Set service consumers to Ports

	// Set CORBA Service Ports

	// </rtc-template>

}

RTC_OmniWheel::~RTC_OmniWheel() {
}

RTC::ReturnCode_t RTC_OmniWheel::onInitialize() {
	// <rtc-template block="bind_config">
	// Bind variables and configuration variable
	bindParameter("odo_id", m_odo_id, "100");
	bindParameter("odo_pos_x", m_odo_pos_x, "0.0");
	bindParameter("odo_pos_y", m_odo_pos_y, "0.0");
	bindParameter("odo_pos_theta", m_odo_pos_theta, "0.0");
	bindParameter("check_ResetPos", m_check_ResetPos, "1");
	bindParameter("check_Log", m_check_Log, "1");
	bindParameter("gain_odo_pos_x", m_gain_odo_pos_x, "1.0");
	bindParameter("gain_odo_pos_y", m_gain_odo_pos_y, "1.0");
	bindParameter("gain_odo_pos_theta", m_gain_odo_pos_theta, "1.0");

	// </rtc-template>

	if (motorctl.Init()) {
		if (m_check_Log)
			printf("MoveWheelChair Init Error\n");
		return RTC::RTC_ERROR;
	}

	//Configurationで自己位置を設定できるようにする。デフォルトは0
	float odo[3];
	odo[0] = m_odo_pos_x;
	odo[1] = m_odo_pos_y;
	odo[2] = m_odo_pos_theta;
	motorctl.SetOdometry(odo);
	active_flag = 1;

	return RTC::RTC_OK;
}

RTC::ReturnCode_t RTC_OmniWheel::onFinalize() {
	motorctl.Stop();
	motorctl.Servo(0);
	motorctl.Close();

#ifdef RECORD
	for(int i=0;i<loop_count;i++){
		period_log << i << " " << sample_time[i] << " " << periodic_time
		[i] << std::endl;
	}
	period_log << "MaxPeriod[us] " << timespec2micro(max_period) << " MinPeriod[us] " << timespec2micro(min_period) << std::endl;
	period_log.close();
	printf("Max Peroid: %ld[us]\n", timespec2micro(max_period));
	printf("Min Peroid: %ld[us]\n", timespec2micro(min_period));
#endif
	return RTC::RTC_OK;
}

/*
 RTC::ReturnCode_t RTC_OmniWheel::onStartup(RTC::UniqueId ec_id)
 {
 return RTC::RTC_OK;
 }
 */
/*
 RTC::ReturnCode_t RTC_OmniWheel::onShutdown(RTC::UniqueId ec_id)
 {
 return RTC::RTC_OK;
 }
 */

RTC::ReturnCode_t RTC_OmniWheel::onActivated(RTC::UniqueId ec_id) {

	//アクティベイト時にFTMDがエラー状態だった場合に備えて一旦サーボOFFする
	motorctl.Stop();
	motorctl.Servo(0);
	usleep(1000);

	//もし、リセットフラグがたっている場合は、アクティベイトの度にConfigurationで自己位置を設定できるようにする。デフォルトは0
	float odo[3];
	if(m_check_ResetPos || active_flag){
		odo[0] = m_odo_pos_x;
		odo[1] = m_odo_pos_y;
		odo[2] = m_odo_pos_theta;
		motorctl.SetOdometry(odo);
	}
	active_flag = 0;

	motorctl.SetOdometryGain(m_gain_odo_pos_x, m_gain_odo_pos_y, m_gain_odo_pos_theta);
	usleep(1000);
	motorctl.Servo(1);
	usleep(1000);

#ifdef RECORD
    first = 1;
    loop_count = 0;
    min_period.tv_sec = 1000;
    min_period.tv_nsec = 1000000;
    timespec_zero(max_period);

    period_log.open("periodic.log");

    periodic_time = new unsigned long [20000];
    sample_time = new unsigned long [20000];

    clock_gettime(CLOCK_REALTIME, &ts_start);
#endif

    if(m_check_Log)
    	printf("onActivated OK\n");

	return RTC::RTC_OK;
}

RTC::ReturnCode_t RTC_OmniWheel::onDeactivated(RTC::UniqueId ec_id) {
	motorctl.Stop();
	motorctl.Servo(0);

#ifdef RECORD
	delete [] sample_time;
	delete [] periodic_time;
#endif

	if(m_check_Log)
		printf("onDeactivated OK\n");

	return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_OmniWheel::onExecute(RTC::UniqueId ec_id) {
	clock_gettime(CLOCK_REALTIME, &ts_now);
#ifdef RECORD
	static struct timespec period, ts_pass;


	sample_time[loop_count] = timespec2micro(ts_now);

	ExecutionContext_var ec = this->get_context(ec_id);

	period = ts_now;

	if(first == 1){
		timespec_sub(period, ts_start);
		first = 0;
	}
	else {
		timespec_sub(period, ts_pass);

		min_period = timespec_min(min_period, period);
		max_period = timespec_max(max_period, period);
		periodic_time[loop_count] = timespec2micro(period);
	}

	ts_pass = ts_now;
	loop_count++;

	if(loop_count > 10000){
		std::cout << "trans to deactivated" << std::endl;
		ec->deactivate_component(this->getObjRef());
	}
#endif

	static float vx=0.0f, vy=0.0f, vt=0.0f;

	if(m_inVelIISIn.isNew()){
		m_inVelIISIn.read();
		vx = (float)m_inVelIIS.vx;
		vy = (float)m_inVelIIS.vy;
		vt = (float)m_inVelIIS.w;
	}
	if(motorctl.Move(vx, vy, vt)){
		if(m_check_Log)
			std::cout << "ERROR::Move" << std::endl;
		return RTC::RTC_ERROR;
	}

	float odometry[3];
	motorctl.GetOdometry(odometry);

	m_outOdometryIIS.x = odometry[0];
	m_outOdometryIIS.y = odometry[1];
	m_outOdometryIIS.theta = odometry[2];
	m_outOdometryIIS.tm.sec = ts_now.tv_sec;
	m_outOdometryIIS.tm.nsec = ts_now.tv_nsec;

	m_outOdometryIIS.id.length(2);
	m_outOdometryIIS.id[0] = m_odo_id;
	m_outOdometryIISOut.write();

	if(m_check_Log == 1){
		std::cout << " in.vx:";
		std::cout << std::setw(8) << std::showpos << std::showpoint << std::right;
		std::cout << m_inVelIIS.vx << "[m/s]";

		std::cout << " in.vy:";
		std::cout << std::setw(8) << std::showpos << std::showpoint << std::right;
		std::cout << m_inVelIIS.vy << "[m/s]";

		std::cout << " in.w:";
		std::cout << std::setw(8) << std::showpos << std::showpoint << std::right;
		std::cout << m_inVelIIS.w << "[rad/s]" << std::endl;

//		std::cout << "tv_sec:" << ts_now.tv_sec << " tv_nsec:" << ts_now.tv_nsec << std::endl;
		std::cout << " out.odo.id:";
		std::cout << std::noshowpos << std::noshowpoint;
		std::cout << m_outOdometryIIS.id[0];

		std::cout << " out.odo.x:";
		std::cout << std::setw(8) << std::showpos << std::showpoint << std::right;
		std::cout << m_outOdometryIIS.x << "[m]";

		std::cout << " out.odo.y:";
		std::cout << std::setw(8) << std::showpos << std::showpoint << std::right;
		std::cout << m_outOdometryIIS.y << "[m]";

		std::cout << " out.odo.theta:";
		std::cout << std::setw(8) << std::showpos << std::showpoint << std::right;
		std::cout << m_outOdometryIIS.theta << "[rad] ";

		std::cout << std::setw(8) << std::showpos << std::showpoint << std::right;
		std::cout << m_outOdometryIIS.theta * 180/M_PI << "[degree]";

		std::cout << std::endl;
	}
	return RTC::RTC_OK;
}


/*
 RTC::ReturnCode_t RTC_OmniWheel::onAborting(RTC::UniqueId ec_id)
 {
 return RTC::RTC_OK;
 }
 */
/*
 RTC::ReturnCode_t RTC_OmniWheel::onError(RTC::UniqueId ec_id)
 {
 return RTC::RTC_OK;
 }
 */
/*
 RTC::ReturnCode_t RTC_OmniWheel::onReset(RTC::UniqueId ec_id)
 {
 return RTC::RTC_OK;
 }
 */
/*
 RTC::ReturnCode_t RTC_OmniWheel::onStateUpdate(RTC::UniqueId ec_id)
 {
 return RTC::RTC_OK;
 }
 */
/*
 RTC::ReturnCode_t RTC_OmniWheel::onRateChanged(RTC::UniqueId ec_id)
 {
 return RTC::RTC_OK;
 }
 */

extern "C" {

	void RTC_OmniWheelInit(RTC::Manager* manager) {
		coil::Properties profile(rtc_omniwheel_spec);
		manager->registerFactory(profile,
				RTC::Create<RTC_OmniWheel>,
				RTC::Delete<RTC_OmniWheel>);
	}

}
;

